/**
 *
 * $Id: DragC.c,v 1.18 1997/02/13 04:02:42 miers Exp $
 *
 * Copyright (C) 1995 Free Software Foundation, Inc.
 *
 * This file is part of the GNU LessTif Library.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **/

/* This code owes a lot to Daniel Dardailler, who wrote up an
   initial implementation of the dynamic drag protocol for an X/Open
   contract while still with the X consortium.
   
   Here's his copyright notice:

   Copyright 1996 Daniel Dardailler.  
   Permission to use, copy, modify, distribute, and sell this software
   for any purpose is hereby granted without fee, provided that the above
   copyright notice appear in all copies and that both that copyright
   notice and this permission notice appear in supporting documentation,
   and that the name of Daniel Dardailler not be used in advertising or
   publicity pertaining to distribution of the software without specific,
   written prior permission.  Daniel Dardailler makes no representations
   about the suitability of this software for any purpose.  It is
   provided "as is" without express or implied warranty.
   */

static char rcsid[] = "$Id: DragC.c,v 1.18 1997/02/13 04:02:42 miers Exp $";

#include <LTconfig.h>
#include <XmI/XmI.h>

#include <Xm/XmP.h>
#include <Xm/DragCP.h>
#include <Xm/DropSMgrP.h>
#include <XmI/DragDropI.h>
#include <Xm/DisplayP.h>
#include <Xm/AtomMgr.h>
#include <Xm/MenuUtilP.h> /* for the _XmGrab stuff */
#include <X11/Xfuncs.h>

#include <XmI/DebugUtil.h>

/* Forward Declarations */

static void class_initialize();
static void class_part_initialize(WidgetClass w_class);
static void initialize(Widget request, Widget new_w, ArgList args, Cardinal *num_args);
static void destroy(Widget w);

static Boolean set_values(Widget current, Widget request, Widget new_w, ArgList args, Cardinal *num_args);

static void _XmDCStartDrag(XmDragContext dc,
			   Widget srcW,
			   XEvent *event);
static void _XmDCCancelDrag(XmDragContext dc);

unsigned char _XmByteOrder(void);
static void internal_notify(Widget w, XtPointer cd, XtPointer cbs);
static void external_notify(Widget w, XtPointer cd, XtPointer cbs);

#define Offset(field) XtOffsetOf(XmDragContextRec, drag.field)

/* Resources for the dragContext class */

static XtResource resources[] = {
    {
	XmNsourceWidget, XmCSourceWidget, XmRWidget,
	sizeof(Widget), Offset(sourceWidget),
	XmRImmediate, (XtPointer)NULL
    },
    {
	XmNexportTargets, XmCExportTargets, XmRAtomList,
	sizeof(Atom *), Offset(exportTargets),
	XmRImmediate, NULL
    },
    {
	XmNnumExportTargets, XmCNumExportTargets, XmRInt,
	sizeof(Cardinal), Offset(numExportTargets),
	XmRImmediate, (XtPointer)0
    },
    {
	XmNconvertProc, XmCConvertProc, XmRFunction,
	sizeof(XmConvertSelectionRec), Offset(convertProc),
	XmRImmediate, (XtPointer)NULL
    },
    {
	XmNclientData, XmCClientData, XtRPointer,
	sizeof(XtPointer), Offset(clientData),
	XmRImmediate, (XtPointer)NULL
    },
    {
	XmNincremental, XmCIncremental, XmRBoolean,
	sizeof(Boolean), Offset(incremental),
	XmRImmediate, (XtPointer)False
    },
    {
	XmNdragOperations, XmCDragOperations, XmRUnsignedChar,
	sizeof(unsigned char), Offset(dragOperations),
	XmRImmediate, (XtPointer)(XmDROP_COPY | XmDROP_MOVE)
    },
    {
	XmNsourceCursorIcon, XmCSourceCursorIcon, XmRWidget,
	sizeof(Widget), Offset(sourceCursorIcon),
	XmRImmediate, (XtPointer)NULL
    },
    {
	XmNsourcePixmapIcon, XmCSourcePixmapIcon, XmRWidget,
	sizeof(Widget), Offset(sourcePixmapIcon),
	XmRImmediate, (XtPointer)NULL
    },
    {
	XmNstateCursorIcon, XmCStateCursorIcon, XmRWidget,
	sizeof(Widget), Offset(stateCursorIcon),
	XmRImmediate, (XtPointer)NULL
    },
    {
	XmNoperationCursorIcon, XmCOperationCursorIcon, XmRWidget,
	sizeof(Widget), Offset(operationCursorIcon),
	XmRImmediate, (XtPointer)NULL
    },
    {
	XmNcursorBackground, XmCCursorBackground, XmRPixel,
	sizeof(Pixel), Offset(cursorBackground),
	XmRCallProc, (XtPointer)_XmBackgroundColorDefault
    },
    {
	XmNcursorForeground, XmCCursorForeground, XmRPixel,
	sizeof(Pixel), Offset(cursorForeground),
	XmRCallProc, (XtPointer)_XmForegroundColorDefault
    },
    {
	XmNvalidCursorForeground, XmCValidCursorForeground, XmRPixel,
	sizeof(Pixel), Offset(validCursorForeground),
	XtRCallProc, (XtPointer)_XmForegroundColorDefault
    },
    {
	XmNinvalidCursorForeground, XmCInvalidCursorForeground, XmRPixel,
	sizeof(Pixel), Offset(invalidCursorForeground),
	XmRCallProc, (XtPointer)_XmForegroundColorDefault
    },
    {
	XmNnoneCursorForeground, XmCNoneCursorForeground, XmRPixel,
	sizeof(Pixel), Offset(noneCursorForeground),
	XmRCallProc, (XtPointer)_XmForegroundColorDefault
    },
    {
	XmNdropSiteEnterCallback, XmCCallback, XmRCallback,
	sizeof(XtCallbackList), Offset(siteEnterCallback),
	XmRImmediate, (XtPointer)NULL
    },
    {
	XmNdropSiteLeaveCallback, XmCCallback, XmRCallback,
	sizeof(XtCallbackList), Offset(siteLeaveCallback),
	XmRImmediate, (XtPointer)NULL
    },
    {
	XmNtopLevelEnterCallback, XmCCallback, XmRCallback,
	sizeof(XtCallbackList), Offset(topLevelEnterCallback),
	XmRImmediate, (XtPointer)NULL
    },
    {
	XmNdragMotionCallback, XmCCallback, XmRCallback,
	sizeof(XtCallbackList), Offset(dragMotionCallback),
	XmRImmediate, (XtPointer)NULL
    },
    {
	XmNtopLevelLeaveCallback, XmCCallback, XmRCallback,
	sizeof(XtCallbackList), Offset(topLevelLeaveCallback),
	XmRImmediate, (XtPointer)NULL
    },
    {
	XmNdropStartCallback, XmCCallback, XmRCallback,
	sizeof(XtCallbackList), Offset(dropStartCallback),
	XmRImmediate, (XtPointer)NULL
    },
    {
	XmNdragDropFinishCallback, XmCCallback, XmRCallback,
	sizeof(XtCallbackList), Offset(dragDropFinishCallback),
	XmRImmediate, (XtPointer)NULL
    },
    {
	XmNdropFinishCallback, XmCCallback, XmRCallback,
	sizeof(XtCallbackList), Offset(dropFinishCallback),
	XmRImmediate, (XtPointer)NULL
    },
    {
	XmNoperationChangedCallback, XmCCallback, XmRCallback,
	sizeof(XtCallbackList), Offset(operationChangedCallback),
	XmRImmediate, (XtPointer)NULL
    },
    {
	XmNblendModel, XmCBlendModel, XmRBlendModel,
	sizeof(unsigned char), Offset(blendModel),
	XtRImmediate, (XtPointer)XmBLEND_ALL
    },
    {
	XmNsourceIsExternal, XmCSourceIsExternal, XmRBoolean,
	sizeof(Boolean), Offset(sourceIsExternal),
	XtRImmediate, (XtPointer)False
    },
    {
	XmNsourceWindow, XmCSourceWindow, XmRWindow,
	sizeof(Window), Offset(srcWindow),
	XtRImmediate, (XtPointer)0
    },
    {
	XmNstartTime, XmCStartTime, XmRInt,
	sizeof(Time), Offset(dragStartTime),
	XtRImmediate, (XtPointer)0
    },
    {
	XmNiccHandle, XmCICCHandle, XmRAtom,
	sizeof(Atom), Offset(iccHandle),
	XtRImmediate, (XtPointer)0
    }
};

static void _XmDCCancel(Widget w, XEvent *event, String *params, Cardinal *num_params);
static void _XmDCMotion(Widget w, XEvent *event, String *params, Cardinal *num_params);
static void _XmDCFinish(Widget w, XEvent *event, String *params, Cardinal *num_params);
static void _XmDCHelp(Widget w, XEvent *event, String *params, Cardinal *num_params);
static void _XmDCIgnore(Widget w, XEvent *event, String *params, Cardinal *num_params);

char _XmDCTranslations[] =
   "<Btn2Up>:             FinishDrag()\n\
    <Btn1Up>:             FinishDrag()\n\
    <BtnUp>:              IgnoreButtons()\n\
    <BtnDown>:            IgnoreButtons()\n\
    <Key>osfCancel:       CancelDrag()\n\
    <Key>osfHelp:         HelpDrag()\n\
    Button2<Enter>:       DragMotion()\n\
    Button2<Leave>:       DragMotion()\n\
    Button2<Motion>:      DragMotion()\n\
    Button1<Enter>:       DragMotion()\n\
    Button1<Leave>:       DragMotion()\n\
    Button1<Motion>:      DragMotion()";

static XtActionsRec actions[] = {
    {"CancelDrag",    _XmDCCancel},
    {"DragMotion",    _XmDCMotion},
    {"FinishDrag",    _XmDCFinish},
    {"HelpDrag",      _XmDCHelp},
    {"IgnoreButtons", _XmDCIgnore}
};

static XmBaseClassExtRec _XmDragCCoreClassExtRec = {
    /* next_extension            */ NULL,
    /* record_type               */ NULLQUARK,                             
    /* version                   */ XmBaseClassExtVersion,
    /* size                      */ sizeof(XmBaseClassExtRec),
    /* initialize_prehook        */ NULL, /* FIXME */
    /* set_values_prehook        */ NULL, /* FIXME */
    /* initialize_posthook       */ NULL, /* FIXME */
    /* set_values_posthook       */ NULL, /* FIXME */
    /* secondary_object_class    */ NULL, /* FIXME */
    /* secondary_object_create   */ NULL, /* FIXME */
    /* get_secondary_resources   */ NULL, /* FIXME */
    /* fast_subclass             */ { 0 }, /* FIXME */
    /* get_values_prehook        */ NULL, /* FIXME */
    /* get_values_posthook       */ NULL, /* FIXME */
    /* class_part_init_prehook   */ NULL, /* FIXME */
    /* class_part_init_posthook  */ NULL, /* FIXME */
    /* ext_resources             */ NULL, /* FIXME */
    /* compiled_ext_resources    */ NULL, /* FIXME */
    /* num_ext_resources         */ 0, /* FIXME */
    /* use_sub_resources         */ False, /* FIXME */
    /* widget_navigable          */ NULL,
    /* focus_change              */ NULL,
    /* wrapper_data              */ NULL
};

XmDragContextClassRec xmDragContextClassRec = {
    /* Core class part */
    {
	/* superclass            */ (WidgetClass) &coreClassRec,
        /* class_name            */ "XmDragContext",
	/* widget_size           */ sizeof(XmDragContextRec),
	/* class_initialize      */ class_initialize,
	/* class_part_initialize */ class_part_initialize,
	/* class_inited          */ False,
	/* initialize            */ initialize,
	/* initialize_hook       */ NULL,
	/* realize               */ XtInheritRealize,
	/* actions               */ actions,
	/* num_actions           */ XtNumber(actions),
	/* resources             */ resources,
	/* num_resources         */ XtNumber(resources),
	/* xrm_class             */ NULLQUARK,
	/* compress_motion       */ True,
	/* compress_exposure     */ XtExposeCompressMaximal,
	/* compress_enterleave   */ True,
	/* visible_interest      */ False,
	/* destroy               */ destroy,
	/* resize                */ NULL,
	/* expose                */ NULL,
	/* set_values            */ set_values,
	/* set_values_hook       */ NULL,
	/* set_values_almost     */ XtInheritSetValuesAlmost,
	/* get_values_hook       */ NULL,
	/* accept_focus          */ NULL,
	/* version               */ XtVersion,
	/* callback offsets      */ NULL,
	/* tm_table              */ _XmDCTranslations,
	/* query_geometry        */ NULL,
	/* display_accelerator   */ XtInheritDisplayAccelerator,
	/* extension             */ (XtPointer)&_XmDragCCoreClassExtRec
    },
    /* DragContext Class part */
    {
        /* start     */ _XmDCStartDrag,
        /* cancel    */ _XmDCCancelDrag,
        /* extension */ NULL,
    }
};

WidgetClass xmDragContextClass = (WidgetClass)&xmDragContextClassRec;

static void
class_initialize()
{
    _XmDragCCoreClassExtRec.record_type = XmQmotif;
}

static void
class_part_initialize(WidgetClass widget_class)
{
    WidgetClass sc;

    _XmFastSubclassInit(widget_class, XmDRAG_CONTEXT_BIT);

    sc = widget_class->core_class.superclass;

    if (DCC_DragStartProc(widget_class) == XmInheritDragStartProc)
	DCC_DragStartProc(widget_class) = DCC_DragStartProc(sc);

    if (DCC_DragCancelProc(widget_class) == XmInheritDragCancelProc)
	DCC_DragCancelProc(widget_class) = DCC_DragCancelProc(sc);
}

static void
initialize(Widget request,
	   Widget new_w,
	   ArgList args,
	   Cardinal *num_args)
{
    Atom *exports;
    XmDropSiteManagerObject dsm;
    Arg argl[2];

    XdbDebug(__FILE__, new_w, "Initialize\n");

    DC_InDropSite(new_w) = False;
    DC_DragTimerId(new_w) = None;

    DC_RoundOffTime(new_w) = 32;
    DC_DragFinishTime(new_w) = CurrentTime;	/* read 0 */
    DC_DropFinishTime(new_w) = CurrentTime;

    /* copy from resource */
    DC_ActiveBlendModel(new_w) = DC_BlendModel(new_w);

    DC_StartX(new_w) = DC_StartY(new_w) = 0;

    /* this appears to be the case, but I have no clue why */
    if (XmIsMotifWMRunning(XtParent(new_w)))
	DC_TrackingMode(new_w) = False;
    else
	DC_TrackingMode(new_w) = True;

    /* no cursor yet */
    DC_CurDragOver(new_w) = NULL;
    DC_OrigDragOver(new_w) = NULL;

    /* setup the notify cb */
    dsm = _XmGetDropSiteManagerObject(_XmDCtoDD(new_w));
    XtSetArg(argl[0], XmNclientData, (XtPointer)new_w);
    if (DC_SourceIsExternal(new_w))
	XtSetArg(argl[1], XmNnotifyProc, external_notify);
    else
	XtSetArg(argl[1], XmNnotifyProc, internal_notify);
    XtSetValues((Widget)dsm, args, 2);

    /* copy the export targets */
    if (DC_ExportTargets(new_w)) {
	exports = (Atom *)XtMalloc(DC_NumExportTargets(new_w) * sizeof(Atom));

	bcopy(DC_ExportTargets(new_w), exports,
	      DC_NumExportTargets(new_w) * sizeof(Atom));

	DC_ExportTargets(new_w) = exports;
    }

    XtWidth(new_w) = XtHeight(new_w) = 32;
    XtX(new_w) = XtY(new_w) = 0;

    /* Setup the receiver info to known state */
    DC_CurrReceiverInfo(new_w) = NULL;
    DC_ReceiverInfos(new_w) = NULL;
    DC_NumReceiverInfos(new_w) = 0;
    DC_MaxReceiverInfos(new_w) = 0;

    XtRealizeWidget(new_w);

    XdbDebug(__FILE__, new_w, "Initialize: Parent %s window 0x%X us 0x%X\n", 
	XtName(XtParent(new_w)), XtWindow(XtParent(new_w)), XtWindow(new_w));
}

static void
destroy(Widget w)
{
    XdbDebug(__FILE__, w, "Destroy\n");

    if (DC_ExportTargets(w))
	XtFree((char *)DC_ExportTargets(w));
}

static Boolean
set_values(Widget old,
	   Widget request,
	   Widget new_w,
	   ArgList args,
	   Cardinal *num_args)
{
    XdbDebug(__FILE__, new_w, "SetValues\n");
    return False;
}

static void
internal_notify(Widget w, XtPointer cd, XtPointer cbs)
{
}

static void
external_notify(Widget w, XtPointer cd, XtPointer cbs)
{
}

static Boolean
drop_convert_incr_callback(Widget w, Atom *selection, Atom *target,
			   Atom *type, XtPointer *val,
			   unsigned long *length, int *format,
			   unsigned long *maxlen, XtPointer cd,
			   XtRequestId *receiver)
{
    return False;
}

static void
drop_lose_incr_callback(Widget w, Atom *selection, XtPointer cd)
{
}

static Boolean
drop_convert_callback(Widget w, Atom *selection, Atom *target,
		      Atom *type, XtPointer *val,
		      unsigned long *length, int *format)
{
    return False;
}

static void
drop_lose_callback(Widget w, Atom *selection)
{
}

static void
new_screen(Widget dc, Display *display)
{
    DC_CurDragOver(dc) =
	(XmDragOverShellWidget)XtCreatePopupShell("cursor",
						  xmDragOverShellWidgetClass,
						  dc, NULL, 0);
}

static void 
_XmDCStartDrag(XmDragContext dc,
	       Widget srcW,
	       XEvent *event)
{
    unsigned int state;
    Time time;
    XmDisplay disp;

    XdbDebug(__FILE__, srcW, "_XmDCStartDrag\n");

    disp = _XmDCtoDD((Widget)dc);

    state = event->xbutton.state;
    time = event->xbutton.time;

    Display_ActiveDC(disp) = dc;
    Display_UserGrabbed(disp) = True;

    DC_DragStartTime(dc) = time;
    DC_LastChangeTime(dc) = time;
    DC_CrossingTime(dc) = time;

    XtX(dc) = DC_StartX(dc) = event->xbutton.x_root;
    XtY(dc) = DC_StartY(dc) = event->xbutton.y_root;

    DC_CurDragOver(dc) = NULL;
    DC_OrigDragOver(dc) = NULL;

    /* find the shell this guy lives in */
    DC_SrcShell(dc) = _XmFindTopMostShell(srcW);
    DC_SrcWindow(dc) = XtWindow(DC_SrcShell(dc));

    DC_ICCHandle(dc) = _XmAllocMotifAtom(DC_SrcShell(dc), DC_DragStartTime(dc));

    if (DC_Incremental(dc))
	XtOwnSelectionIncremental(DC_SrcShell(dc), DC_ICCHandle(dc),
				  DC_DragStartTime(dc), 
				  drop_convert_incr_callback,
				  drop_lose_incr_callback,
				  NULL,	NULL, DC_ClientData(dc));
    else
	XtOwnSelection(DC_SrcShell(dc), DC_ICCHandle(dc),
		       DC_DragStartTime(dc),
		       drop_convert_callback,
		       drop_lose_callback,
		       NULL);

    DC_SourceIsExternal(dc) = False;
    DC_ServerGrabbed(dc) = False;

    DC_ActiveProtocolStyle(dc) = _XmGetActiveProtocolStyle((Widget)dc);

    /* the protocol gotten from _XmGetActiveProtocolStyle() needs to be
     * filtered for reasonable values for the initiator. */
    switch (DC_ActiveProtocolStyle(dc)) {
	case XmDRAG_DROP_ONLY:
	    DC_ActiveProtocolStyle(dc) = XmDRAG_NONE;
	    break;

	case XmDRAG_PREREGISTER:
	    DC_ActiveProtocolStyle(dc) = XmDRAG_DYNAMIC;
	    break;

	case XmDRAG_NONE:
	case XmDRAG_PREFER_PREREGISTER:
	case XmDRAG_PREFER_DYNAMIC:
	case XmDRAG_DYNAMIC:
	default:
	    break;
    }

    DC_Operations(dc) = DC_DragOperations(dc);
    DC_LastEventState(dc) = state;

    /* fool around with mask */
    if (DC_DragOperations(dc) & XmDROP_MOVE) {
	DC_Operation(dc) = DC_DragOperations(dc) & XmDROP_MOVE;
	DC_Operations(dc) = DC_DragOperations(dc) & XmDROP_MOVE;
    }
    else if (DC_DragOperations(dc) & XmDROP_COPY) {
	DC_Operation(dc) = DC_DragOperations(dc) & XmDROP_COPY;
	DC_Operations(dc) = DC_DragOperations(dc) & XmDROP_COPY;
    }
    else if (DC_DragOperations(dc) & XmDROP_LINK) {
	DC_Operation(dc) = DC_DragOperations(dc) & XmDROP_LINK;
	DC_Operations(dc) = DC_DragOperations(dc) & XmDROP_LINK;
    }

    new_screen((Widget)dc, XtDisplay(dc));

    DC_SourceWidget(dc) = srcW;

    XGrabPointer(XtDisplay(dc), DefaultRootWindow(XtDisplay(dc)), True,
		 ButtonPressMask | ButtonMotionMask | ButtonReleaseMask |
		 EnterWindowMask | LeaveWindowMask,
		 GrabModeSync, GrabModeAsync,
		 None, _XmDragOverGetActiveCursor((Widget)DC_CurDragOver(dc)),
		 DC_DragStartTime(dc));

    XtAddGrab((Widget)dc, True, True);

    XAllowEvents(XtDisplay(dc), SyncPointer, DC_LastChangeTime(dc));
}

static void 
_XmDCCancelDrag(XmDragContext dc)
{
    XdbDebug(__FILE__, NULL, "_XmDCCancelDrag\n");

    XUngrabPointer(XtDisplay(dc), DC_LastChangeTime(dc));
    XtRemoveGrab((Widget)dc);
}

static void
_XmDCCancel(Widget w,
	    XEvent *event,
	    String *params,
	    Cardinal *num_params)
{
    XdbDebug(__FILE__, w, "_XmDCCancel\n");

    DC_LastChangeTime(w) = event->xbutton.time;

    /* needs to free the ICCHandle */
    XUngrabPointer(XtDisplay(w), DC_LastChangeTime(w));
    XtRemoveGrab((Widget)w);
}

static void 
_XmDCMotion(Widget w,
	    XEvent *event,
	    String *params,
	    Cardinal *num_params)
{
    if (event->type == MotionNotify)
	XdbDebug(__FILE__, w, "_XmDCMotion: Motion\n");
    else if (event->type == EnterNotify)
	XdbDebug(__FILE__, w, "_XmDCMotion: Enter\n");
    else if (event->type == LeaveNotify)
	XdbDebug(__FILE__, w, "_XmDCMotion: Leave\n");

    DC_LastChangeTime(w) = event->xbutton.time;
}

static void 
_XmDCFinish(Widget w,
	    XEvent *event,
	    String *params,
	    Cardinal *num_params)
{
    XdbDebug(__FILE__, w, "_XmDCFinish\n");

    /* needs to free the ICCHandle */

    DC_DragFinishTime(w) = event->xbutton.time;
    DC_LastChangeTime(w) = DC_DragFinishTime(w);

    XUngrabPointer(XtDisplay(w), DC_LastChangeTime(w));
    XtRemoveGrab(w);
}

static void 
_XmDCHelp(Widget w,
	  XEvent *event,
	  String *params,
	  Cardinal *num_params)
{
    XdbDebug(__FILE__, w, "_XmDCHelp\n");

    DC_LastChangeTime(w) = event->xbutton.time;
}

/* 
 * We have to allowevents here or else if someone is dragging with
 * button two down and hits button one, we freeze.
 */
static void 
_XmDCIgnore(Widget w,
	    XEvent *event,
	    String *params,
	    Cardinal *num_params)
{
    XdbDebug(__FILE__, w, "_XmDCIgnore\n");

    XAllowEvents(XtDisplay(w), SyncPointer, event->xbutton.time);
}

XmDragReceiverInfo
_XmAllocReceiverInfo(XmDragContext dc)
{
    XdbDebug(__FILE__, NULL, "_XmAllocReceiverInfo\n");
    return NULL;
}

/*
 * array is dragInitiator (7) by dragReceiver (6).  Matrix comes from
 * 1-371 in OSF/Motif Programmer's Reference
 * Order is:
 *  XmDRAG_NONE,
 *  XmDRAG_DROP_ONLY,
 *  XmDRAG_PREFER_PREREGISTER,
 *  XmDRAG_PREREGISTER,
 *  XmDRAG_PREFER_DYNAMIC,
 *  XmDRAG_DYNAMIC,
 *  XmDRAG_PREFER_RECEIVER
 */
static unsigned char
protocolStyle[XmDRAG_PREFER_RECEIVER+1][XmDRAG_DYNAMIC+1] = {
    /* dragInitiator is XmDRAG_NONE */
    { XmDRAG_NONE,               XmDRAG_NONE,              XmDRAG_NONE,
      XmDRAG_NONE,               XmDRAG_NONE,              XmDRAG_NONE },

    /* dragInitiator is XmDRAG_DROP_ONLY */
    { XmDRAG_NONE,               XmDRAG_DROP_ONLY,         XmDRAG_DROP_ONLY,
      XmDRAG_DROP_ONLY,          XmDRAG_DROP_ONLY,         XmDRAG_DROP_ONLY },

    /* dragInitiator is XmDRAG_PREFER_PREREGISTER */
    { XmDRAG_NONE,               XmDRAG_DROP_ONLY,         XmDRAG_PREREGISTER,
      XmDRAG_PREREGISTER,        XmDRAG_PREREGISTER,       XmDRAG_DYNAMIC },

    /* dragInitiator is XmDRAG_PREREGISTER */
    { XmDRAG_NONE,               XmDRAG_DROP_ONLY,         XmDRAG_PREREGISTER,
      XmDRAG_PREREGISTER,        XmDRAG_PREREGISTER,       XmDRAG_DROP_ONLY },

    /* dragInitiator is XmDRAG_PREFER_DYNAMIC */
    { XmDRAG_NONE,               XmDRAG_DROP_ONLY,         XmDRAG_DYNAMIC,
      XmDRAG_PREREGISTER,        XmDRAG_DYNAMIC,           XmDRAG_DYNAMIC },

    /* dragInitiator is XmDRAG_DYNAMIC */
    { XmDRAG_NONE,               XmDRAG_DROP_ONLY,         XmDRAG_DYNAMIC,
      XmDRAG_DROP_ONLY,          XmDRAG_DYNAMIC,           XmDRAG_DYNAMIC },

    /* dragInitiator is XmDRAG_PREFER_RECEIVER */
    { XmDRAG_NONE,               XmDRAG_DROP_ONLY,         XmDRAG_PREREGISTER,
      XmDRAG_PREREGISTER,        XmDRAG_DYNAMIC,           XmDRAG_DYNAMIC },
};

unsigned char
_XmGetActiveProtocolStyle(Widget w)
{
    unsigned char dis, drs;

    XdbDebug(__FILE__, w, "_XmGetActiveProtocolStyle\n");

    dis = Display_DragInitiatorProtocolStyle(_XmDCtoDD(w));
    drs = Display_DragReceiverProtocolStyle(_XmDCtoDD(w));

    if (!DC_SourceIsExternal(w)) {

	if (DC_CurrReceiverInfo(w))
	    drs = DC_CurrReceiverInfo(w)->dragProtocolStyle;

	return protocolStyle[dis][drs];
    }
    else {
	switch (drs) {
	case XmDRAG_DROP_ONLY:
	case XmDRAG_PREFER_PREREGISTER:
	case XmDRAG_PREREGISTER:
	case XmDRAG_PREFER_DYNAMIC:
	case XmDRAG_DYNAMIC:
	    return XmDRAG_DYNAMIC;

	case XmDRAG_NONE:
	default:
	    return XmDRAG_NONE;
	}
    }
}

unsigned char 
_XmByteOrder(void)
{
  static unsigned char byte_order = 0;
  
  if (!byte_order) {
    unsigned int endian = 1;

    byte_order = (*((char *)&endian)) ? 'l' : 'B';
  }

  return byte_order;
}
